import 'package:flutter/material.dart';
import 'package:flutter_hooks/flutter_hooks.dart';
import 'package:flutter_query/flutter_query.dart';
import 'package:flutter_query/src/query.dart';
import 'package:provider/provider.dart';

typedef Future<T> QueryFunction<T>();
typedef Future<T> PageableQueryFunction<T>(int page);
typedef dynamic GetFetchMoreFunction<T>(T latestPage);

AsyncSnapshot<T> useQuery<T>(String key, QueryFunction<T> fn) {
  return use(_QueryHook<T>(key, fn));
}

InfiniteQueryContext<T> useInfiniteQuery<T>(
  String key,
  PageableQueryFunction<T> fn, {
  GetFetchMoreFunction<T> getFetchMore,
}) {
  return use(_InfiniteQueryHook(key, fn, getFetchMore));
}

class _QueryHook<T> extends Hook<AsyncSnapshot<T>> {
  const _QueryHook(this.key, this.fn);

  final String key;
  final QueryFunction<T> fn;

  @override
  _QueryStateHook<T> createState() => _QueryStateHook();
}

class _QueryStateHook<T> extends HookState<AsyncSnapshot<T>, _QueryHook<T>> {
  AsyncSnapshot<T> _snapshot;
  Function _watcher;
  QueryCache _cache;

  @override
  void initHook() {
    _snapshot = AsyncSnapshot<T>.withData(ConnectionState.none, null);

    try {
      _cache = Provider.of<QueryCache>(context, listen: false);
      final key = hook.key;
      final query = _cache.registerQuery(key, QueryConfig(fn: hook.fn));
      if (query.data != null) {
        _snapshot = AsyncSnapshot<T>.withData(ConnectionState.done, query.data);
      }
      _watcher = (data) {
        if (data != null) {
          setState(() {
            _snapshot = AsyncSnapshot<T>.withData(ConnectionState.done, data);
          });
        }
      };
      _cache.watchQuery(key, _watcher);
    } catch (e) {
      throw Exception(
        'Cache not found! You either need to use the QueryCacheProvider or provide a query cache yourself.',
      );
    }
    super.initHook();
  }

  @override
  void dispose() {
    _cache.unwatchQuery(hook.key, _watcher);
    super.dispose();
  }

  @override
  AsyncSnapshot<T> build(BuildContext context) {
    return _snapshot;
  }
}

class InfiniteQueryContext<T> {
  AsyncSnapshot<T> snapshot;
  Function fetchMore;
  bool hasMorePages = true;
}

class _InfiniteQueryHook<T> extends Hook<InfiniteQueryContext<T>> {
  const _InfiniteQueryHook(this.key, this.fn, this.getFetchMore);

  final String key;
  final PageableQueryFunction<T> fn;
  final Function getFetchMore;

  @override
  HookState<InfiniteQueryContext<T>, Hook<InfiniteQueryContext<T>>>
      createState() => _InfiniteQueryStateHook();
}

class _InfiniteQueryStateHook<T>
    extends HookState<InfiniteQueryContext<T>, _InfiniteQueryHook<T>> {
  InfiniteQueryContext<T> _context;
  QueryCache _cache;
  Function _watcher;

  @override
  void initHook() {
    _context = InfiniteQueryContext();
    _context.snapshot = AsyncSnapshot<T>.withData(ConnectionState.none, null);
    try {
      _cache = Provider.of<QueryCache>(context, listen: false);
      final key = hook.key;
      _cache.registerQuery(
        key,
        QueryConfig(
          fn: hook.fn,
          getFetchMore: hook.getFetchMore ?? (lastCursor) => lastCursor + 1,
          isInfinite: true,
        ),
      );
      _context.fetchMore = () {
        _cache.invalidateQuery(key);
      };
      _watcher = (data) {
        if (data != null) {
          setState(() {
            if (data is List && data.isEmpty) {
              _context.hasMorePages = false;
            }

            _context.snapshot =
                AsyncSnapshot<T>.withData(ConnectionState.done, data);
          });
        }
      };
      _cache.watchQuery(key, _watcher);
    } catch (e) {
      throw Exception(
        'Cache not found! You either need to use the QueryCacheProvider or provide a query cache yourself.',
      );
    }
    super.initHook();
  }

  @override
  InfiniteQueryContext<T> build(BuildContext context) {
    return _context;
  }
}

// const state = useInfiniteQuery('posts', (page) => http.getPosts(page), getFetchMore: (cursor, lastPage, allPages) => cursor + 1)
// onScrollBottom = () {
// if (!state.snapshot.waiting()) {
//  state.fetchMore()
// }
//}
