import 'dart:async';

import 'package:flutter/material.dart';
import 'package:provider/provider.dart';

class Query<T> {
  final String key;
  final List<Function> observers = [];
  QueryConfig config;
  T data;

  Query(this.key, this.config);

  Future<T> fetch() async {
    Future<T> future;
    if (config.isInfinite) {
      final cursor = config.getFetchMore(data);
      if (cursor != false) {
        future = config.fn(cursor);
      }
    } else {
      future = config.fn();
    }

    return future.then((data) {
      if (config.isInfinite) {
        (this.data as List ?? []).addAll((data as List));
      } else {
        this.data = data;
      }
      return data;
    });
  }
}

class QueryConfig {
  final isInfinite;
  final Function fn;
  final Function getFetchMore;

  QueryConfig({
    @required this.fn,
    this.getFetchMore,
    this.isInfinite = false,
  });
}

class QueryCache {
  Map<String, Query> _queryMap = {};

  Query getQueryByKey(String key) {
    return _queryMap[key];
  }

  Query registerQuery(String key, QueryConfig config) {
    var query = getQueryByKey(key);
    if (query == null) {
      query = Query(key, config);
      _queryMap[key] = query;
      invalidateQuery(key);
    }
    return query;
  }

  void watchQuery(String key, Function observer) {
    var query = getQueryByKey(key);
    if (query != null) {
      query.observers.add(observer);
      if (query.data != null) {
        observer(query.data);
      }
    }
  }

  void unwatchQuery(String key, Function observer) {
    var query = getQueryByKey(key);
    if (query != null) {
      query.observers.remove(observer);
    }
  }

  void invalidateQuery(String key) {
    final query = getQueryByKey(key);
    if (query != null) {
      query.fetch().then((data) {
        query.observers.forEach((observer) => observer(data));
      });
    }
  }
}

class QueryCacheProvider extends StatefulWidget {
  final Widget child;
  QueryCacheProvider({this.child});

  @override
  State<StatefulWidget> createState() => QueryCacheProviderState();
}

class QueryCacheProviderState extends State<QueryCacheProvider> {
  QueryCache _queryCache;

  @override
  void initState() {
    _queryCache = QueryCache();
    super.initState();
  }

  @override
  Widget build(BuildContext context) {
    return Provider.value(
      value: _queryCache,
      child: widget.child,
    );
  }
}
